#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from packaging import version
from argparse import ArgumentParser

import torch
from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence
from transformers import AutoModel
from ltp.nn import (
    SharedDropout, MLP, LSTM, Bilinear, BaseModule, Triaffine,
    MFVISemanticDependency, LBPSemanticDependency,
)
from ltp.transformer_biaffine import GraphResult

torch_version = version.parse(torch.__version__)
pack_padded_sequence_cpu_version = version.parse('1.7.0')


class ViClassifier(nn.Module):
    def __init__(self, input_size, label_num, dropout,
                 lstm_hidden_size=600, lstm_num_layers=3, bin_hidden_size=150,
                 arc_hidden_size=600, rel_hidden_size=600, loss_interpolation=0.4, inference='mfvi', max_iter=3):
        super().__init__()
        self.label_num = label_num
        self.loss_interpolation = loss_interpolation

        if lstm_num_layers > 0:
            self.lstm = LSTM(
                input_size=input_size,
                hidden_size=lstm_hidden_size,
                num_layers=lstm_num_layers,
                bidirectional=True,
                dropout=dropout
            )
            self.lstm_dropout = SharedDropout(p=dropout)
            hidden_size = lstm_hidden_size * 2
        else:
            self.lstm = None
            hidden_size = input_size

        self.mlp_bin_d = MLP([hidden_size, bin_hidden_size], output_dropout=dropout)
        self.mlp_bin_h = MLP([hidden_size, bin_hidden_size], output_dropout=dropout)
        self.mlp_bin_g = MLP([hidden_size, bin_hidden_size], output_dropout=dropout)
        self.mlp_arc_h = MLP([hidden_size, arc_hidden_size], output_dropout=dropout)
        self.mlp_arc_d = MLP([hidden_size, arc_hidden_size], output_dropout=dropout)
        self.mlp_rel_h = MLP([hidden_size, rel_hidden_size], output_dropout=dropout)
        self.mlp_rel_d = MLP([hidden_size, rel_hidden_size], output_dropout=dropout)

        self.sib_attn = Triaffine(bin_hidden_size, bias_x=True, bias_y=True)
        self.cop_attn = Triaffine(bin_hidden_size, bias_x=True, bias_y=True)
        self.grd_attn = Triaffine(bin_hidden_size, bias_x=True, bias_y=True)
        self.arc_atten = Bilinear(arc_hidden_size, arc_hidden_size, 1, bias_x=True, bias_y=True, expand=True)
        self.rel_atten = Bilinear(rel_hidden_size, rel_hidden_size, label_num, bias_x=True, bias_y=True, expand=True)

        self.vi = (MFVISemanticDependency if inference == 'mfvi' else LBPSemanticDependency)(max_iter)

    def forward(self, input, attention_mask=None, word_index=None, word_attention_mask=None, head=None, labels=None,
                is_processed=False):
        if not is_processed:
            assert word_attention_mask is not None
            input = input[:, :-1, :]
            if word_index is not None:
                input = torch.cat([input[:, :1, :], torch.gather(
                    input[:, 1:, :], dim=1, index=word_index.unsqueeze(-1).expand(-1, -1, input.size(-1))
                )], dim=1)

        if self.lstm is not None:
            lengths = word_attention_mask.sum(1) + 1  # +cls
            if torch_version >= pack_padded_sequence_cpu_version:
                lengths = lengths.cpu()
            input = pack_padded_sequence(input, lengths, True, False)
            input, _ = self.lstm(input)
            input, _ = pad_packed_sequence(input, True, total_length=word_attention_mask.shape[1] + 1)
            input = self.lstm_dropout(input)

        bin_d = self.mlp_bin_d(input)
        bin_h = self.mlp_bin_h(input)
        bin_g = self.mlp_bin_g(input)

        arc_h = self.mlp_arc_h(input)
        arc_d = self.mlp_arc_d(input)

        rel_h = self.mlp_rel_h(input)
        rel_d = self.mlp_rel_d(input)

        # [batch_size, seq_len, seq_len, n_labels]
        s_sib = self.sib_attn(bin_d, bin_d, bin_h).triu_()
        s_sib = (s_sib + s_sib.transpose(-1, -2)).permute(0, 3, 1, 2)
        # [batch_size, seq_len, seq_len, n_labels]
        s_cop = self.cop_attn(bin_h, bin_d, bin_h).permute(0, 3, 1, 2).triu_()
        s_cop = s_cop + s_cop.transpose(-1, -2)
        # [batch_size, seq_len, seq_len, n_labels]
        s_grd = self.grd_attn(bin_g, bin_d, bin_h).permute(0, 3, 1, 2)

        # [batch_size, seq_len, seq_len]
        s_arc = self.arc_atten(arc_d, arc_h).squeeze_(1)
        # [batch_size, seq_len, seq_len, n_labels]
        s_rel = self.rel_atten(rel_d, rel_h).permute(0, 2, 3, 1)

        loss = None

        # cat cls
        mask = torch.cat([word_attention_mask[:, :1], word_attention_mask], dim=1)
        mask = mask.unsqueeze(1) & mask.unsqueeze(2)
        mask[:, 0] = 0

        if labels is not None:
            rel_loss = nn.CrossEntropyLoss()

            head = torch.cat([torch.zeros_like(head[:, :1, :], device=head.device), head], dim=1)
            arc_mask = head.gt(0) & mask
            arc_loss, arc_logits = self.vi((s_arc, s_sib, s_cop, s_grd), mask, head)

            labels = torch.cat([torch.zeros_like(labels[:, :1, :], device=labels.device), labels], dim=1)
            rel_loss = rel_loss(s_rel[arc_mask], labels[arc_mask])
            loss = self.loss_interpolation * rel_loss + (1 - self.loss_interpolation) * arc_loss
        else:
            arc_logits = self.vi((s_arc, s_sib, s_cop, s_grd), mask)

        return GraphResult(loss=loss, arc_logits=arc_logits, rel_logits=s_rel, src_arc_logits=s_arc)


class TransformerVi(BaseModule):
    def __init__(self, hparams, config=None):
        super().__init__()
        self.save_hyperparameters(hparams)
        if config is None:
            self.transformer = AutoModel.from_pretrained(self.hparams.transformer)
        else:
            self.transformer = AutoModel.from_config(config)
        self.dropout = nn.Dropout(self.hparams.dropout)
        hidden_size = self.transformer.config.hidden_size
        self.classifier = ViClassifier(
            input_size=hidden_size,
            label_num=self.hparams.num_labels,
            dropout=self.hparams.dropout,
            lstm_num_layers=self.hparams.lstm_num_layers,
            lstm_hidden_size=self.hparams.lstm_hidden_size,
            bin_hidden_size=self.hparams.bin_hidden_size,
            arc_hidden_size=self.hparams.arc_hidden_size,
            rel_hidden_size=self.hparams.rel_hidden_size,
            loss_interpolation=self.hparams.loss_interpolation,
            inference=self.hparams.inference,
            max_iter=self.hparams.max_iter,
        )

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False, conflict_handler='resolve')
        parser.add_argument('--transformer', type=str, default="hfl/chinese-electra-base-discriminator")
        parser.add_argument('--lstm_num_layers', type=int, default=0)
        parser.add_argument('--lstm_hidden_size', type=int, default=600)
        parser.add_argument('--arc_hidden_size', type=int, default=600)
        parser.add_argument('--rel_hidden_size', type=int, default=600)
        parser.add_argument('--bin_hidden_size', type=int, default=150)
        parser.add_argument('--loss_interpolation', type=float, default=0.1)
        parser.add_argument('--dropout', type=float, default=0.1)
        parser.add_argument('--num_labels', type=int)
        parser.add_argument('--inference', type=str, default='mfvi')
        parser.add_argument('--max_iter', type=int, default=3)
        return parser

    def forward(
            self,
            input_ids=None,
            logits_mask=None,
            attention_mask=None,
            word_index=None,
            word_attention_mask=None,
            token_type_ids=None,
            position_ids=None,
            head_mask=None,
            inputs_embeds=None,
            head=None,
            labels=None
    ) -> GraphResult:
        hidden_states = self.transformer(
            input_ids,
            attention_mask,
            token_type_ids,
            position_ids,
            head_mask,
            inputs_embeds,
            output_attentions=False,
            output_hidden_states=False,
            return_dict=False,
        )
        sequence_output = hidden_states[0]
        sequence_output = self.dropout(sequence_output)

        return self.classifier(
            input=sequence_output,
            attention_mask=attention_mask,
            word_index=word_index,
            word_attention_mask=word_attention_mask,
            head=head,
            labels=labels
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import os
from argparse import ArgumentParser, Namespace
from collections import OrderedDict
from packaging.version import parse as version_parse

import torch
from transformers import AutoConfig
from ltp.transformer_multitask import TransformerMultiTask as Model

from ltp import __version__


def load_labels(*paths):
    try:
        labels_path = os.path.join(*paths)
        with open(labels_path, encoding='utf-8') as f:
            return [line.strip() for line in f.readlines()]
    except Exception as e:
        return []


def deploy_model(args, version=__version__):
    parsed_version = version_parse(version)
    assert parsed_version.major == 4

    if parsed_version.minor == 0:
        deploy_model_4_0(args, version)
    elif parsed_version.minor == 1:
        deploy_model_4_1(args, version)


def deploy_model_4_1(args, version):
    from argparse import Namespace
    from pytorch_lightning.core.saving import ModelIO

    fake_parser = ArgumentParser()
    fake_parser = Model.add_model_specific_args(fake_parser)
    hparams = fake_parser.parse_args(args=[])

    try:
        ckpt = torch.load(args.resume_from_checkpoint, map_location='cpu')
    except AttributeError as e:
        if "_gpus_arg_default" in e.args[0]:
            from ltp.patchs import pl_1_2_patch_1_1
            patched_model_path = pl_1_2_patch_1_1(args.resume_from_checkpoint)
            ckpt = torch.load(patched_model_path, map_location='cpu')
        else:
            raise e
    hparams.__dict__.update(ckpt[ModelIO.CHECKPOINT_HYPER_PARAMS_KEY])
    transformer_config = AutoConfig.from_pretrained(hparams.transformer)
    model = Model(hparams, config=transformer_config)
    model.load_state_dict(ckpt['state_dict'])

    model_config = Namespace(**model.hparams)
    # LOAD VOCAB
    pos_labels = load_labels(args.pos_data_dir, 'vocabs', 'xpos.txt')
    ner_labels = load_labels(args.ner_data_dir, 'vocabs', 'bio.txt')
    srl_labels = load_labels(args.srl_data_dir, 'vocabs', 'arguments.txt')
    dep_labels = load_labels(args.dep_data_dir, 'vocabs', 'deprel.txt')
    sdp_labels = load_labels(args.sdp_data_dir, 'vocabs', 'deps.txt')

    # MODEL CLIP
    if not len(pos_labels):
        del model.pos_classifier
        model_config.pos_num_labels = 0

    if not len(ner_labels):
        del model.ner_classifier
        model_config.ner_num_labels = 0

    if not len(srl_labels):
        del model.srl_classifier
        model_config.srl_num_labels = 0

    if not len(dep_labels):
        del model.dep_classifier
        model_config.dep_num_labels = 0

    if not len(sdp_labels):
        del model.sdp_classifier
        model_config.sdp_num_labels = 0

    model_state_dict = OrderedDict(model.state_dict().items())

    ltp_model = {
        'version': version,
        'model': model_state_dict,
        'model_config': model_config,
        'transformer_config': model.transformer.config.to_dict(),
        'seg': ['I-W', 'B-W'],
        'pos': pos_labels,
        'ner': ner_labels,
        'srl': srl_labels,
        'dep': dep_labels,
        'sdp': sdp_labels,
    }
    os.makedirs(args.ltp_model, exist_ok=True)
    torch.save(ltp_model, os.path.join(args.ltp_model, 'ltp.model'))

    from transformers import AutoTokenizer
    tokenizer = AutoTokenizer.from_pretrained(args.transformer)
    tokenizer.save_pretrained(args.ltp_model)


def deploy_model_4_0(args, version):
    ltp_adapter_mapper = sorted([
        ('transformer', 'pretrained'),
        ('seg_classifier', 'seg_decoder'),
        ('pos_classifier', 'pos_decoder'),
        ('ner_classifier', 'ner_decoder'),
        ('ner_classifier.classifier', 'ner_decoder.mlp'),
        ('ner_classifier.relative_transformer', 'ner_decoder.transformer'),
        ('srl_classifier', 'srl_decoder'),
        ('srl_classifier.rel_atten', 'srl_decoder.biaffine'),
        ('srl_classifier.crf', 'srl_decoder.crf'),
        ('dep_classifier', 'dep_decoder'),
        ('sdp_classifier', 'sdp_decoder'),
    ], key=lambda x: len(x[0]), reverse=True)

    model = Model.load_from_checkpoint(
        args.resume_from_checkpoint, hparams=args
    )
    model_state_dict = OrderedDict(model.state_dict().items())
    for preffix, target_preffix in ltp_adapter_mapper:
        model_state_dict = {
            key.replace(preffix, target_preffix, 1): value
            for key, value in model_state_dict.items()
        }

    pos_labels = load_labels(args.pos_data_dir, 'vocabs', 'xpos.txt')
    ner_labels = load_labels(args.ner_data_dir, 'ner_labels.txt')
    srl_labels = load_labels(args.srl_data_dir, 'srl_labels.txt')
    dep_labels = load_labels(args.dep_data_dir, 'vocabs', 'deprel.txt')
    sdp_labels = load_labels(args.sdp_data_dir, 'vocabs', 'deps.txt')

    ltp_model = {
        'version': '4.0.0',
        'code_version': version,
        'seg': ['I-W', 'B-W'],
        'pos': pos_labels,
        'ner': ner_labels,
        'srl': srl_labels,
        'dep': dep_labels,
        'sdp': sdp_labels,
        'pretrained_config': model.transformer.config,
        'model_config': {
            'class': 'SimpleMultiTaskModel',
            'init': {
                'seg': {'label_num': args.seg_num_labels},
                'pos': {'label_num': args.pos_num_labels},
                'ner': {
                    'label_num': args.ner_num_labels,
                    'decoder': 'RelativeTransformer',
                    'RelativeTransformer': {
                        'num_heads': args.ner_num_heads,
                        'num_layers': args.ner_num_layers,
                        'hidden_size': args.ner_hidden_size,
                        'dropout': args.dropout
                    }
                },
                'dep': {
                    'label_num': args.dep_num_labels, 'decoder': 'Graph',
                    'Graph': {
                        'arc_hidden_size': args.dep_arc_hidden_size,
                        'rel_hidden_size': args.dep_rel_hidden_size,
                        'dropout': args.dropout
                    }
                },
                'sdp': {
                    'label_num': args.sdp_num_labels, 'decoder': 'Graph',
                    'Graph': {
                        'arc_hidden_size': args.sdp_arc_hidden_size,
                        'rel_hidden_size': args.sdp_rel_hidden_size,
                        'dropout': args.dropout
                    }
                },
                'srl': {
                    'label_num': args.srl_num_labels, 'decoder': 'BiLinearCRF',
                    'BiLinearCRF': {
                        'hidden_size': args.srl_hidden_size,
                        'dropout': args.dropout
                    }
                }
            }
        },
        'model': model_state_dict
    }
    os.makedirs(args.ltp_model, exist_ok=True)
    torch.save(ltp_model, os.path.join(args.ltp_model, 'ltp.model'))

    from transformers import AutoTokenizer
    tokenizer = AutoTokenizer.from_pretrained(args.transformer)
    tokenizer.save_pretrained(args.ltp_model)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import os
from argparse import ArgumentParser

import numpy
import torch
import torch.utils.data
from tqdm import tqdm
from pytorch_lightning import Trainer

from ltp import optimization
from ltp.data import dataset as datasets
from ltp.data.utils import collate
from ltp.metrics.metric import Seqeval
from ltp.task_part_of_speech import tokenize
from ltp.transformer_rel_linear import TransformerRelLinear as Model


from ltp.utils import TaskInfo, common_train, map2device, convert2npy, tune_train, dataset_cache_wrapper, \
    add_common_specific_args
from ltp.utils import add_tune_specific_args

os.environ['TOKENIZERS_PARALLELISM'] = 'true'


# CUDA_VISIBLE_DEVICES=0 PYTHONPATH=. python ltp/task_named_entity_recognition.py --data_dir=data/ner --num_labels=13 --max_epochs=10 --batch_size=16 --gpus=1 --precision=16
@dataset_cache_wrapper(
    extra_builder=lambda dataset: Seqeval(dataset[datasets.Split.TRAIN].features['labels'].feature.names)
)
def build_dataset(data_dir, task_name, tokenizer, max_length=512, **kwargs):
    dataset = datasets.load_dataset(
        datasets.Bio,
        data_dir=data_dir,
        cache_dir=data_dir,
        data_files=datasets.Bio.default_files(data_dir)
    )
    dataset.rename_column_('bio', 'labels')
    dataset = dataset.map(
        lambda examples: tokenize(examples, tokenizer, max_length), batched=True,
        cache_file_names={
            k: d._get_cache_file_path(f"{task_name}-{k}-tokenized") for k, d in dataset.items()
        }
    )
    dataset = dataset.filter(
        lambda x: not x['overflow'],
        cache_file_names={
            k: d._get_cache_file_path(f"{task_name}-{k}-filtered") for k, d in dataset.items()
        }
    )
    dataset.set_format(type='torch', columns=[
        'input_ids', 'token_type_ids', 'attention_mask', 'word_index', 'word_attention_mask', 'labels'
    ])
    return dataset


def validation_method(metric: Seqeval, task=f'ner', preffix='val', ret=False):
    def step(self: Model, batch, batch_nb):
        result = self.forward(**batch)

        if 'word_attention_mask' in batch:
            mask = batch['word_attention_mask']
        else:
            mask = batch['attention_mask'][:, 2:] == 1

        if result.decoded is None:
            # acc
            preds = torch.argmax(result.logits, dim=-1)
            step_result = metric.step(batch, preds, batch['labels'], mask)
        else:
            preds = result.decoded
            labels = result.labels
            labels[~mask] = -1
            labels = [[word for word in sent if word != -1] for sent in result.labels.cpu().numpy()]
            step_result = metric.step(batch, preds, labels, mask)

        step_result['loss'] = result.loss.item()
        return step_result

    def epoch_end(self: Model, outputs):
        if isinstance(outputs, dict):
            outputs = [outputs]
        length = len(outputs)
        loss = sum([output['loss'] for output in outputs]) / length

        core_metric, epoch_result = metric.epoch_end(outputs)
        dictionary = {f'{task}/{preffix}_{k}': v for k, v in epoch_result.items()}
        dictionary[f'{task}/{preffix}_loss'] = loss

        self.log_dict(
            dictionary=dictionary,
            on_step=False, on_epoch=True, prog_bar=True, logger=True
        )
        if ret:
            return core_metric

    return step, epoch_end


task_info = TaskInfo(
    task_name='ner',
    metric_name='f1',
    build_dataset=build_dataset,
    validation_method=validation_method
)


def add_task_specific_args(parent_parser):
    parser = ArgumentParser(parents=[parent_parser], add_help=False)
    parser.add_argument('--batch_size', type=int, default=16)
    parser.add_argument('--num_workers', type=int, default=4)
    parser.add_argument('--data_dir', type=str, required=True)
    parser.add_argument('--build_dataset', action='store_true')
    return parser


def build_distill_dataset(args):
    model = Model.load_from_checkpoint(
        args.resume_from_checkpoint, hparams=args
    )

    model.eval()
    model.freeze()

    dataset, metric = build_dataset(args.data_dir, task_info.task_name, model)
    train_dataloader = torch.utils.data.DataLoader(
        dataset[datasets.Split.TRAIN],
        batch_size=args.batch_size,
        collate_fn=collate,
        num_workers=args.num_workers
    )

    output = os.path.join(args.data_dir, task_info.task_name, 'output.npz')

    if torch.cuda.is_available():
        model.cuda()
        map2cpu = lambda x: map2device(x)
        map2cuda = lambda x: map2device(x, model.device)
    else:
        map2cpu = lambda x: x
        map2cuda = lambda x: x

    with torch.no_grad():
        batchs = []
        for batch in tqdm(train_dataloader):
            batch = map2cuda(batch)
            logits = model.forward(**batch).logits
            batch.update(logits=logits)
            batchs.append(map2cpu(batch))

        try:
            numpy.savez(
                output,
                data=convert2npy(batchs),
                extra=convert2npy({
                    'transitions': model.classifier.crf.transitions,
                    'start_transitions': model.classifier.crf.start_transitions,
                    'end_transitions': model.classifier.crf.end_transitions
                })
            )
        except Exception as e:
            numpy.savez(output, data=convert2npy(batchs))

    print("Done")


def main():
    parser = ArgumentParser()

    # add task level args
    parser = add_common_specific_args(parser)
    parser = add_tune_specific_args(parser)
    parser = add_task_specific_args(parser)

    # add model specific args
    parser = Model.add_model_specific_args(parser)
    parser = optimization.add_optimizer_specific_args(parser)
    parser = Trainer.add_argparse_args(parser)

    # set default args
    parser.set_defaults(gradient_clip_val=1.0, min_epochs=1, max_epochs=10)
    parser.set_defaults(num_labels=13)

    args = parser.parse_args()

    if args.build_dataset:
        build_distill_dataset(args)
    elif args.tune:
        tune_train(args, model_class=Model, task_info=task_info)
    else:
        common_train(args, model_class=Model, task_info=task_info)


if __name__ == '__main__':
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from .collate import collate
from .iterator import iter_raw_lines, iter_lines, iter_blocks
from .multitask_dataloader import MultiTaskDataloader
from .vocab_helper import vocab_builder



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from . import utils

try:
    from . import dataset
except Exception as e:
    from types import ModuleType


    class _Dataset(ModuleType):
        def __init__(self):
            super().__init__("dataset")

        def __getattr__(self, item):
            print("Need Install datasets!!!")
            print("pip install datasets")


    dataset = _Dataset()



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from packaging import version

_patch_version = version.parse('4.1.3')


def patch(ckpt):
    if 'version' in ckpt and version.parse(ckpt['version']) < _patch_version:
        srl_decoder_mlp_rel_h = {}
        srl_decoder_mlp_rel_d = {}

        if ckpt['seg'][0].startswith('B'):
            ckpt['seg'] = list(reversed(ckpt['seg']))
            seg_classifier_weight = ckpt['model']['seg_classifier.weight']
            ckpt['model']['seg_classifier.weight'] = seg_classifier_weight[[1, 0]]

        for key, value in ckpt['model'].items():
            key: str
            if key.startswith('srl_classifier.mlp_rel_h'):
                srl_decoder_mlp_rel_h[key] = value
            elif key.startswith('srl_classifier.mlp_rel_d'):
                srl_decoder_mlp_rel_d[key] = value
        srl_decoder_mlp_rel_h = {
            k.replace('srl_classifier.mlp_rel_h', 'srl_classifier.mlp_rel_d'): v for k, v in
            srl_decoder_mlp_rel_h.items()
        }
        srl_decoder_mlp_rel_d = {
            k.replace('srl_classifier.mlp_rel_d', 'srl_classifier.mlp_rel_h'): v for k, v in
            srl_decoder_mlp_rel_d.items()
        }

        try:
            ckpt['model']['srl_classifier.crf.start_transitions'] \
                = ckpt['model'].pop('srl_classifier.rel_crf.start_transitions')
            ckpt['model']['srl_classifier.crf.end_transitions'] \
                = ckpt['model'].pop('srl_classifier.rel_crf.end_transitions')
            ckpt['model']['srl_classifier.crf.transitions'] \
                = ckpt['model'].pop('srl_classifier.rel_crf.transitions')
        except Exception as e:
            pass

        ckpt['model'].update(srl_decoder_mlp_rel_h)
        ckpt['model'].update(srl_decoder_mlp_rel_d)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from .eisner import eisner
from .sent_split import split_sentence
from .maximum_forward_matching import Trie



# -*- coding: utf-8 -*-

import torch
import torch.nn as nn


class SharedDropout(nn.Module):
    r"""
    SharedDropout differs from the vanilla dropout strategy in that
    the dropout mask is shared across one dimension.
    Args:
        p (float):
            The probability of an element to be zeroed. Default: 0.5.
        batch_first (bool):
            If ``True``, the input and output tensors are provided as ``[batch_size, seq_len, *]``.
            Default: ``True``.
    Examples:
        >>> x = torch.ones(1, 3, 5)
        >>> nn.Dropout()(x)
        tensor([[[0., 2., 2., 0., 0.],
                 [2., 2., 0., 2., 2.],
                 [2., 2., 2., 2., 0.]]])
        >>> SharedDropout()(x)
        tensor([[[2., 0., 2., 0., 2.],
                 [2., 0., 2., 0., 2.],
                 [2., 0., 2., 0., 2.]]])
    """

    def __init__(self, p=0.5, batch_first=True):
        super().__init__()

        self.p = p
        self.batch_first = batch_first

    def __repr__(self):
        s = f"p={self.p}"
        if self.batch_first:
            s += f", batch_first={self.batch_first}"

        return f"{self.__class__.__name__}({s})"

    def forward(self, x):
        r"""
        Args:
            x (~torch.Tensor):
                A tensor of any shape.
        Returns:
            The returned tensor is of the same shape as `x`.
        """

        if self.training:
            if self.batch_first:
                mask = self.get_mask(x[:, 0], self.p).unsqueeze(1)
            else:
                mask = self.get_mask(x[0], self.p)
            x = x * mask

        return x

    @staticmethod
    def get_mask(x, p):
        return x.new_empty(x.shape).bernoulli_(1 - p) / (1 - p)


class IndependentDropout(nn.Module):
    r"""
    For :math:`N` tensors, they use different dropout masks respectively.
    When :math:`N-M` of them are dropped, the remaining :math:`M` ones are scaled by a factor of :math:`N/M` to compensate,
    and when all of them are dropped together, zeros are returned.
    Args:
        p (float):
            The probability of an element to be zeroed. Default: 0.5.
    Examples:
        >>> x, y = torch.ones(1, 3, 5), torch.ones(1, 3, 5)
        >>> x, y = IndependentDropout()(x, y)
        >>> x
        tensor([[[1., 1., 1., 1., 1.],
                 [0., 0., 0., 0., 0.],
                 [2., 2., 2., 2., 2.]]])
        >>> y
        tensor([[[1., 1., 1., 1., 1.],
                 [2., 2., 2., 2., 2.],
                 [0., 0., 0., 0., 0.]]])
    """

    def __init__(self, p=0.5):
        super().__init__()

        self.p = p

    def __repr__(self):
        return f"{self.__class__.__name__}(p={self.p})"

    def forward(self, *items):
        r"""
        Args:
            items (list[~torch.Tensor]):
                A list of tensors that have the same shape except the last dimension.
        Returns:
            The returned tensors are of the same shape as `items`.
        """

        if self.training:
            masks = [x.new_empty(x.shape[:2]).bernoulli_(1 - self.p) for x in items]
            total = sum(masks)
            scale = len(items) / total.max(torch.ones_like(total))
            masks = [mask * scale for mask in masks]
            items = [item * mask.unsqueeze(-1) for item, mask in zip(items, masks)]

        return items


# -*- coding: utf-8 -*-

import torch
import torch.nn as nn


class LBPSemanticDependency(nn.Module):
    r"""
    Loopy Belief Propagation for approximately calculating marginals of semantic dependency trees.
    References:
        - Xinyu Wang, Jingxian Huang and Kewei Tu. 2019.
          `Second-Order Semantic Dependency Parsing with End-to-End Neural Networks`_.
    .. _Second-Order Semantic Dependency Parsing with End-to-End Neural Networks:
        https://www.aclweb.org/anthology/P19-1454/
    """

    def __init__(self, max_iter=3):
        super().__init__()

        self.max_iter = max_iter

    def __repr__(self):
        return f"{self.__class__.__name__}(max_iter={self.max_iter})"

    @torch.enable_grad()
    def forward(self, scores, mask, target=None):
        r"""
        Args:
            scores (~torch.Tensor, ~torch.Tensor):
                Tuple of four tensors `s_edge`, `s_sib`, `s_cop` and `s_grd`.
                `s_edge` (``[batch_size, seq_len, seq_len]``) holds Scores of all possible dependent-head pairs.
                `s_sib` (``[batch_size, seq_len, seq_len, seq_len]``) holds the scores of dependent-head-sibling triples.
                `s_cop` (``[batch_size, seq_len, seq_len, seq_len]``) holds the scores of dependent-head-coparent triples.
                `s_grd` (``[batch_size, seq_len, seq_len, seq_len]``) holds the scores of dependent-head-grandparent triples.
            mask (~torch.BoolTensor): ``[batch_size, seq_len, seq_len]``.
                The mask to avoid aggregation on padding tokens.
            target (~torch.LongTensor): ``[batch_size, seq_len, seq_len]``.
                A Tensor of gold-standard dependent-head pairs. Default: ``None``.
        Returns:
            ~torch.Tensor, ~torch.Tensor:
                The first is the training loss averaged by the number of tokens, which won't be returned if ``target=None``.
                The second is a tensor for marginals of shape ``[batch_size, seq_len, seq_len]``.
        """

        logQ = self.belief_propagation(*(s.requires_grad_() for s in scores), mask)
        marginals = logQ.exp()

        if target is None:
            return marginals
        loss = -logQ.gather(-1, target.unsqueeze(-1))[mask].sum() / mask.sum()

        return loss, marginals

    def belief_propagation(self, s_edge, s_sib, s_cop, s_grd, mask):
        batch_size, seq_len, _ = mask.shape
        hs, ms = torch.stack(torch.where(torch.ones_like(mask[0]))).view(-1, seq_len, seq_len).sort(0)[0].unbind()
        # [seq_len, seq_len, batch_size], (h->m)
        mask = mask.permute(2, 1, 0)
        # [seq_len, seq_len, seq_len, batch_size], (h->m->s)
        mask2o = mask.unsqueeze(1) & mask.unsqueeze(2)
        mask2o = mask2o & hs.unsqueeze(-1).ne(hs.new_tensor(range(seq_len))).unsqueeze(-1)
        mask2o = mask2o & ms.unsqueeze(-1).ne(ms.new_tensor(range(seq_len))).unsqueeze(-1)
        # log potentials of unary and binary factors
        # [2, seq_len, seq_len, batch_size], (h->m)
        p_edge = torch.stack((torch.zeros_like(s_edge), s_edge)).permute(0, 3, 2, 1)
        # [seq_len, seq_len, seq_len, batch_size], (h->m->s)
        p_sib = s_sib.permute(2, 1, 3, 0)
        # [seq_len, seq_len, seq_len, batch_size], (h->m->c)
        p_cop = s_cop.permute(2, 1, 3, 0)
        # [seq_len, seq_len, seq_len, batch_size], (h->m->g)
        p_grd = s_grd.permute(2, 1, 3, 0)

        # log beliefs
        # [2, seq_len, seq_len, batch_size], (h->m)
        b = p_edge.new_zeros(2, seq_len, seq_len, batch_size)
        # log messages of siblings
        # [2, seq_len, seq_len, seq_len, batch_size], (h->m->s)
        m_sib = p_sib.new_zeros(2, seq_len, seq_len, seq_len, batch_size)
        # log messages of co-parents
        # [2, seq_len, seq_len, seq_len, batch_size], (h->m->c)
        m_cop = p_cop.new_zeros(2, seq_len, seq_len, seq_len, batch_size)
        # log messages of grand-parents
        # [2, seq_len, seq_len, seq_len, batch_size], (h->m->g)
        m_grd = p_grd.new_zeros(2, seq_len, seq_len, seq_len, batch_size)

        for _ in range(self.max_iter):
            b = b.log_softmax(0)
            # m(ik->ij) = logsumexp(b(ik) - m(ij->ik) + p(ij->ik)) - m(ik->)
            m = b.unsqueeze(3) - m_sib
            m_sib = torch.stack((m.logsumexp(0), torch.stack((m[0], m[1] + p_sib)).logsumexp(0)))
            m_sib = m_sib.transpose(2, 3).log_softmax(0)
            m = b.unsqueeze(3) - m_cop
            m_cop = torch.stack((m.logsumexp(0), torch.stack((m[0], m[1] + p_cop)).logsumexp(0)))
            m_cop = m_cop.transpose(2, 3).log_softmax(0)
            m = b.unsqueeze(3) - m_grd
            m_grd = torch.stack((m.logsumexp(0), torch.stack((m[0], m[1] + p_grd)).logsumexp(0)))
            m_grd = m_grd.transpose(2, 3).log_softmax(0)
            # b(ij) = p(ij) + sum(m(ik->ij)), min(i, j) < k < max(i, j)
            b = p_edge + ((m_sib + m_cop + m_grd) * mask2o).sum(3)

        return b.permute(3, 2, 1, 0).log_softmax(-1)


class MFVISemanticDependency(nn.Module):
    r"""
    Mean Field Variational Inference for approximately calculating marginals of semantic dependency trees.
    References:
        - Xinyu Wang, Jingxian Huang and Kewei Tu. 2019.
          `Second-Order Semantic Dependency Parsing with End-to-End Neural Networks`_.
    .. _Second-Order Semantic Dependency Parsing with End-to-End Neural Networks:
        https://www.aclweb.org/anthology/P19-1454/
    """

    def __init__(self, max_iter=3):
        super().__init__()

        self.max_iter = max_iter

    def __repr__(self):
        return f"{self.__class__.__name__}(max_iter={self.max_iter})"

    @torch.enable_grad()
    def forward(self, scores, mask, target=None):
        r"""
        Args:
            scores (~torch.Tensor, ~torch.Tensor):
                Tuple of four tensors `s_edge`, `s_sib`, `s_cop` and `s_grd`.
                `s_edge` (``[batch_size, seq_len, seq_len]``) holds Scores of all possible dependent-head pairs.
                `s_sib` (``[batch_size, seq_len, seq_len, seq_len]``) holds the scores of dependent-head-sibling triples.
                `s_cop` (``[batch_size, seq_len, seq_len, seq_len]``) holds the scores of dependent-head-coparent triples.
                `s_grd` (``[batch_size, seq_len, seq_len, seq_len]``) holds the scores of dependent-head-grandparent triples.
            mask (~torch.BoolTensor): ``[batch_size, seq_len, seq_len]``.
                The mask to avoid aggregation on padding tokens.
            target (~torch.LongTensor): ``[batch_size, seq_len, seq_len]``.
                A Tensor of gold-standard dependent-head pairs. Default: ``None``.
        Returns:
            ~torch.Tensor, ~torch.Tensor:
                The first is the training loss averaged by the number of tokens, which won't be returned if ``target=None``.
                The second is a tensor for marginals of shape ``[batch_size, seq_len, seq_len]``.
        """

        logQ = self.mean_field_variational_inference(*(s.requires_grad_() for s in scores), mask)
        marginals = logQ.exp()

        if target is None:
            return marginals
        loss = -logQ.gather(-1, target.unsqueeze(-1))[mask].sum() / mask.sum()

        return loss, marginals

    def mean_field_variational_inference(self, s_edge, s_sib, s_cop, s_grd, mask):
        batch_size, seq_len, _ = mask.shape
        hs, ms = torch.stack(torch.where(torch.ones_like(mask[0]))).view(-1, seq_len, seq_len).sort(0)[0].unbind()
        # [seq_len, seq_len, batch_size], (h->m)
        mask = mask.permute(2, 1, 0)
        # [seq_len, seq_len, seq_len, batch_size], (h->m->s)
        mask2o = mask.unsqueeze(1) & mask.unsqueeze(2)
        mask2o = mask2o & hs.unsqueeze(-1).ne(hs.new_tensor(range(seq_len))).unsqueeze(-1)
        mask2o = mask2o & ms.unsqueeze(-1).ne(ms.new_tensor(range(seq_len))).unsqueeze(-1)
        # [seq_len, seq_len, batch_size], (h->m)
        s_edge = s_edge.permute(2, 1, 0)
        # [seq_len, seq_len, seq_len, batch_size], (h->m->s)
        s_sib = s_sib.permute(2, 1, 3, 0) * mask2o
        # [seq_len, seq_len, seq_len, batch_size], (h->m->c)
        s_cop = s_cop.permute(2, 1, 3, 0) * mask2o
        # [seq_len, seq_len, seq_len, batch_size], (h->m->g)
        s_grd = s_grd.permute(2, 1, 3, 0) * mask2o

        # posterior distributions
        # [2, seq_len, seq_len, batch_size], (h->m)
        q = s_edge.new_zeros(2, seq_len, seq_len, batch_size)

        for _ in range(self.max_iter):
            q = q.softmax(0)
            # f(ij) = sum(q(ik)s^sib(ij,ik) + q(kj)s^cop(ij,kj) + q(jk)s^grd(ij,jk)), k != i,j
            f = (q[1].unsqueeze(1) * s_sib + q[1].transpose(0, 1).unsqueeze(0) * s_cop + q[1].unsqueeze(0) * s_grd).sum(
                2)
            # q(ij) = s(ij) + f(ij)
            q = torch.stack((torch.zeros_like(q[0]), s_edge + f))

        return q.permute(3, 2, 1, 0).log_softmax(-1)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from collections import namedtuple

TaskInfo = namedtuple(
    "TaskInfo",
    [
        "task_name",
        "metric_name",
        "build_dataset",
        "validation_method"
    ]
)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>


from ltp.patchs.pytorch_lightning.patch_1_1_to_1_2 import patch as pl_1_2_patch_1_1

from ltp.patchs.model.patch_4_1_3 import patch as model_patch_4_1_3



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import warnings


def get_entities(seq, suffix=False):
    """Gets entities from sequence.

    Args:
        seq (list): sequence of labels.

    Returns:
        list: list of (chunk_type, chunk_start, chunk_end).

    Example:
        >>> from seqeval.metrics.sequence_labeling import get_entities
        >>> seq = ['B-PER', 'I-PER', 'O', 'B-LOC']
        >>> get_entities(seq)
        [('PER', 0, 1), ('LOC', 3, 3)]
    """

    def _validate_chunk(chunk, suffix):
        if chunk in ['O', 'B', 'I', 'E', 'S']:
            return

        if suffix:
            if not chunk.endswith(('-B', '-I', '-E', '-S')):
                warnings.warn('{} seems not to be NE tag.'.format(chunk))

        else:
            if not chunk.startswith(('B-', 'I-', 'E-', 'S-')):
                warnings.warn('{} seems not to be NE tag.'.format(chunk))

    # for nested list
    if any(isinstance(s, list) for s in seq):
        seq = [item for sublist in seq for item in sublist + ['O']]

    prev_tag = 'O'
    prev_type = ''
    begin_offset = 0
    chunks = []
    for i, chunk in enumerate(seq + ['O']):
        _validate_chunk(chunk, suffix)

        if suffix:
            tag = chunk[-1]
            type_ = chunk[:-1].rsplit('-', maxsplit=1)[0] or '_'
        else:
            tag = chunk[0]
            type_ = chunk[1:].split('-', maxsplit=1)[-1] or '_'

        if end_of_chunk(prev_tag, tag, prev_type, type_):
            chunks.append((prev_type, begin_offset, i - 1))
        if start_of_chunk(prev_tag, tag, prev_type, type_):
            begin_offset = i
        prev_tag = tag
        prev_type = type_

    return chunks


def end_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk ended between the previous and current word.

    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.

    Returns:
        chunk_end: boolean.
    """
    chunk_end = False

    if prev_tag == 'E':
        chunk_end = True
    if prev_tag == 'S':
        chunk_end = True

    if prev_tag == 'B' and tag == 'B':
        chunk_end = True
    if prev_tag == 'B' and tag == 'S':
        chunk_end = True
    if prev_tag == 'B' and tag == 'O':
        chunk_end = True
    if prev_tag == 'I' and tag == 'B':
        chunk_end = True
    if prev_tag == 'I' and tag == 'S':
        chunk_end = True
    if prev_tag == 'I' and tag == 'O':
        chunk_end = True

    if prev_tag != 'O' and prev_tag != '.' and prev_type != type_:
        chunk_end = True

    return chunk_end


def start_of_chunk(prev_tag, tag, prev_type, type_):
    """Checks if a chunk started between the previous and current word.

    Args:
        prev_tag: previous chunk tag.
        tag: current chunk tag.
        prev_type: previous type.
        type_: current type.

    Returns:
        chunk_start: boolean.
    """
    chunk_start = False

    if tag == 'B':
        chunk_start = True
    if tag == 'S':
        chunk_start = True

    if prev_tag == 'E' and tag == 'E':
        chunk_start = True
    if prev_tag == 'E' and tag == 'I':
        chunk_start = True
    if prev_tag == 'S' and tag == 'E':
        chunk_start = True
    if prev_tag == 'S' and tag == 'I':
        chunk_start = True
    if prev_tag == 'O' and tag == 'E':
        chunk_start = True
    if prev_tag == 'O' and tag == 'I':
        chunk_start = True

    if tag != 'O' and tag != '.' and prev_type != type_:
        chunk_start = True

    return chunk_start



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from .bio import Bio
from .conllu import Conllu
from .srl import Srl

from typing import Optional, Union, Dict, List
from datasets import DatasetBuilder, Features, Split, DatasetDict, Dataset
from datasets import Sequence, ClassLabel, Value, Translation, TranslationVariableLanguages


def load_dataset(
        builder_cls: type,
        name: Optional[str] = None,
        data_dir: Optional[str] = None,
        data_files: Union[Dict, List] = None,
        split: Optional[Union[str, Split]] = None,
        cache_dir: Optional[str] = None,
        features: Optional[Features] = None,
        save_infos: bool = False,
        **config_kwargs,
) -> Union[DatasetDict, Dataset]:
    # Instantiate the dataset builder
    builder_instance: DatasetBuilder = builder_cls(
        cache_dir=cache_dir,
        name=name,
        data_dir=data_dir,
        data_files=data_files,
        hash=hash,
        features=features,
        **config_kwargs,
    )

    # Download and prepare data
    builder_instance.download_and_prepare()

    # Build dataset for splits
    ds = builder_instance.as_dataset(split=split)
    if save_infos:
        builder_instance._save_infos()

    return ds



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from torch.utils.data import Dataset, DataLoader


class BatchedDataset(Dataset):
    def __init__(self, batches):
        self.batches = batches

    def __getitem__(self, index):
        return self.batches[index]

    def __len__(self):
        return len(self.batches)


class BatchedDataloader(DataLoader):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import logging

import os
import itertools
from collections import Counter

import datasets
from os.path import join
from dataclasses import dataclass
from ltp.data.utils import iter_blocks, vocab_builder

_CITATION = """\
@misc{nivre2020universal,
    title={Universal Dependencies v2: An Evergrowing Multilingual Treebank Collection},
    author={Joakim Nivre and Marie-Catherine de Marneffe and Filip Ginter and Jan Hajič and Christopher D. Manning and Sampo Pyysalo and Sebastian Schuster and Francis Tyers and Daniel Zeman},
    year={2020},
    eprint={2004.10643},
    archivePrefix={arXiv},
    primaryClass={cs.CL}
}
"""

_DESCRIPTION = """\
Universal Dependencies (UD) is a framework for consistent annotation of grammar (parts of speech, morphological 
features, and syntactic dependencies) across different human languages. UD is an open community effort with over 
300 contributors producing more than 150 treebanks in 90 languages.
"""

_TRAINING_FILE = "train.conllu"
_DEV_FILE = "dev.conllu"
_TEST_FILE = "test.conllu"


@vocab_builder
def build_vocabs(data_dir, *files, min_freq=5):
    counters = {
        'word': (1, Counter()), 'lemma': (2, Counter()), 'upos': (3, Counter()),
        'xpos': (4, Counter()), 'feats': (5, Counter()), 'deprel': (7, Counter()),
        # FOR CHAR FEATS
        'word_char': (1, Counter()),
        # DEPS
        'deps': (8, Counter())
    }

    if any([os.path.exists(os.path.join(data_dir, 'vocabs', f'{key}.txt')) for key in counters]):
        return

    if not os.path.exists(os.path.join(data_dir, 'vocabs')):
        os.makedirs(os.path.join(data_dir, 'vocabs'))

    for filename in files:
        for line_num, block in iter_blocks(filename=filename):
            values = [list(value) for value in zip(*block)]

            for name, (row, counter) in counters.items():
                if 'char' in name:
                    counter.update(itertools.chain(*values[row]))
                elif 'deps' == name:
                    try:
                        deps = [[label.split(':', maxsplit=1)[1] for label in dep.split('|')] for dep in values[row]]
                        counter.update(itertools.chain(*deps))
                    except:
                        counter.update('_')
                else:
                    counter.update(values[row])

    for feat, (row, counter) in counters.items():
        if 'word' in feat:
            counter = Counter({word: count for word, count in counter.items() if count > min_freq})

        with open(os.path.join(data_dir, 'vocabs', f'{feat}.txt'), mode='w') as f:
            f.write('\n'.join(sorted(counter.keys())))


def create_feature(file=None):
    if file:
        return datasets.ClassLabel(names_file=file)
    return datasets.Value('string')


@dataclass
class ConlluConfig(datasets.BuilderConfig):
    """BuilderConfig for Conllu"""

    upos: str = None
    xpos: str = None
    feats: str = None
    deprel: str = None
    deps: str = None


# ["id", "from", "lemma", "upos", "xpos", "feats", "head", "deprel", "deps", "misc"]
class Conllu(datasets.GeneratorBasedBuilder):
    BUILDER_CONFIG_CLASS = ConlluConfig

    @staticmethod
    def default_files(data_dir) -> dict:
        return {
            datasets.Split.TRAIN: join(data_dir, _TRAINING_FILE),
            datasets.Split.VALIDATION: join(data_dir, _DEV_FILE),
            datasets.Split.TEST: join(data_dir, _TEST_FILE),
        }

    def _info(self):
        build_vocabs(self.config)
        feats = {
            'upos': self.config.upos,
            'xpos': self.config.xpos,
            'feats': self.config.feats,
            'deprel': self.config.deprel,
            'deps': self.config.deps
        }

        for key in feats:
            if feats[key] is None:
                feats[key] = os.path.join(self.config.data_dir, 'vocabs', f'{key}.txt')

        deps_rel_feature = create_feature(feats['deps'])
        if deps_rel_feature.num_classes > 1:
            self.config.deps = feats['deps']

        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "id": datasets.Sequence(datasets.Value("int64")),
                    "form": datasets.Sequence(datasets.Value("string")),
                    "lemma": datasets.Sequence(datasets.Value("string")),
                    "upos": datasets.Sequence(create_feature(feats['upos'])),
                    "xpos": datasets.Sequence(create_feature(feats['xpos'])),
                    "feats": datasets.Sequence(datasets.Value("string")),
                    "head": datasets.Sequence(datasets.Value("int64")),
                    "deprel": datasets.Sequence(create_feature(feats['deprel'])),
                    "deps": datasets.Sequence(
                        {
                            'id': datasets.Value('int64'),
                            'head': datasets.Value("int64"),
                            'rel': deps_rel_feature
                        } if deps_rel_feature.num_classes > 1 else create_feature(None)
                    ),
                    "misc": datasets.Sequence(datasets.Value("string")),
                }
            ),
            supervised_keys=None,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        """We handle string, list and dicts in datafiles"""
        if not self.config.data_files:
            raise ValueError(f"At least one data file must be specified, but got data_files={self.config.data_files}")
        data_files = dl_manager.download_and_extract(self.config.data_files)
        if isinstance(data_files, (str, list, tuple)):
            files = data_files
            if isinstance(files, str):
                files = [files]
            return [datasets.SplitGenerator(name=datasets.Split.TRAIN, gen_kwargs={"files": files})]
        splits = []
        for split_name, files in data_files.items():
            if isinstance(files, str):
                files = [files]
            splits.append(datasets.SplitGenerator(name=split_name, gen_kwargs={"files": files}))
        return splits

    def _generate_examples(self, files):
        for filename in files:
            logging.info("⏳ Generating examples from = %s", filename)
            for line_num, block in iter_blocks(filename=filename):
                # last example
                id, words, lemma, upos, xpos, feats, head, deprel, deps, misc = [list(value) for value in zip(*block)]
                if self.config.deps:
                    deps = [[label.split(':', maxsplit=1) for label in dep.split('|')] for dep in deps]
                    deps = [[{'id': depid, 'head': int(label[0]), 'rel': label[-1]} for label in dep] for depid, dep in
                            enumerate(deps)]
                    deps = list(itertools.chain(*deps))
                    if any([dep['head'] >= len(words) for dep in deps]):
                        continue

                yield line_num, {
                    "id": id, "form": words, "lemma": lemma, "upos": upos, "xpos": xpos,
                    "feats": feats, "head": head, "deprel": deprel, "deps": deps, "misc": misc,
                }



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from torch import nn


def MLP(layer_sizes, dropout=0.0, activation=nn.ReLU, output_dropout=None, output_activation=None):
    layers = []
    num_layers = len(layer_sizes) - 1
    for index in range(num_layers):
        if index < num_layers:
            layers.extend([
                nn.Linear(layer_sizes[index], layer_sizes[index + 1]),
                activation(),
                nn.Dropout(p=dropout)
            ])
        else:
            layers.append(nn.Linear(layer_sizes[index], layer_sizes[index + 1]))

            if output_activation is not None:
                layers.append(output_activation())
            if output_activation is not None:
                layers.append(nn.Dropout(p=output_dropout))
    return nn.Sequential(*layers)
